From 0848ae05d4d1e1c2fbefaa5a43a4910efe9bfe98 Mon Sep 17 00:00:00 2001
From: Tomas Vondra <tomas.vondra@postgresql.org>
Date: Sun, 15 Aug 2021 13:23:13 +0200
Subject: [PATCH 3/4] Modify selectivity estimation for Var op Var

This modifies estimation of (Var op Var) clauses with distinct Vars from
the same relation. We recognize the clauses as compatible with extended
statistics, and estimate them using the extended MCV etc.
---
 src/backend/optimizer/path/clausesel.c        |  37 +++-
 src/backend/statistics/extended_stats.c       |  83 +++++--
 src/backend/statistics/mcv.c                  | 172 +++++++++++----
 .../statistics/extended_stats_internal.h      |   4 +-
 src/test/regress/expected/stats_ext.out       | 208 +++++++++---------
 5 files changed, 323 insertions(+), 181 deletions(-)

diff --git a/src/backend/optimizer/path/clausesel.c b/src/backend/optimizer/path/clausesel.c
index 06f836308d0..90ee73f89df 100644
--- a/src/backend/optimizer/path/clausesel.c
+++ b/src/backend/optimizer/path/clausesel.c
@@ -714,6 +714,7 @@ clause_selectivity_ext(PlannerInfo *root,
 	Selectivity s1 = 0.5;		/* default for any unhandled clause type */
 	RestrictInfo *rinfo = NULL;
 	bool		cacheable = false;
+	Node	   *src = clause;
 
 	if (clause == NULL)			/* can this still happen? */
 		return s1;
@@ -871,11 +872,37 @@ clause_selectivity_ext(PlannerInfo *root,
 		}
 		else
 		{
-			/* Estimate selectivity for a restriction clause. */
-			s1 = restriction_selectivity(root, opno,
-										 opclause->args,
-										 opclause->inputcollid,
-										 varRelid);
+			/*
+			 * It might be a single (Expr op Expr) clause, which goes here due
+			 * to the optimization at the beginning of clauselist_selectivity.
+			 * So we try applying extended stats first, and then fall back to
+			 * restriction_selectivity.
+			 */
+			bool	estimated = false;
+
+			if (use_extended_stats)
+			{
+				List	   *clauses = list_make1(src);
+				RelOptInfo *rel = find_single_rel_for_clauses(root, clauses);
+
+				if (rel && rel->rtekind == RTE_RELATION && rel->statlist != NIL)
+				{
+					Bitmapset  *estimatedclauses = NULL;
+
+					s1 = statext_clauselist_selectivity(root, clauses, varRelid,
+														jointype, sjinfo, rel,
+														&estimatedclauses, false);
+
+					estimated = (bms_num_members(estimatedclauses) == 1);
+				}
+			}
+
+			/* Estimate selectivity for a restriction clause (fallback). */
+			if (!estimated)
+				s1 = restriction_selectivity(root, opno,
+											 opclause->args,
+											 opclause->inputcollid,
+											 varRelid);
 		}
 
 		/*
diff --git a/src/backend/statistics/extended_stats.c b/src/backend/statistics/extended_stats.c
index d2aa8d0ca3f..7bfde068aff 100644
--- a/src/backend/statistics/extended_stats.c
+++ b/src/backend/statistics/extended_stats.c
@@ -1352,19 +1352,27 @@ statext_is_compatible_clause_internal(PlannerInfo *root, Node *clause,
 		return true;
 	}
 
-	/* (Var/Expr op Const) or (Const op Var/Expr) */
+	/*
+	 * Three opclause variants are supported: (Expr op Const), (Const op Expr),
+	 * (Expr op Expr). That means we may need to analyze one or two expressions
+	 * to make sure the opclause is compatible with extended stats.
+	 */
 	if (is_opclause(clause))
 	{
 		RangeTblEntry *rte = root->simple_rte_array[relid];
 		OpExpr	   *expr = (OpExpr *) clause;
-		Node	   *clause_expr;
+		ListCell   *lc;
+		List	   *clause_exprs;
 
 		/* Only expressions with two arguments are considered compatible. */
 		if (list_length(expr->args) != 2)
 			return false;
 
-		/* Check if the expression has the right shape */
-		if (!examine_opclause_args(expr->args, &clause_expr, NULL, NULL))
+		/*
+		 * Check if the expression has the right shape. This returns either one
+		 * or two expressions, depending on whether there is a Const.
+		 */
+		if (!examine_opclause_args(expr->args, &clause_exprs, NULL, NULL))
 			return false;
 
 		/*
@@ -1404,13 +1412,31 @@ statext_is_compatible_clause_internal(PlannerInfo *root, Node *clause,
 			!get_func_leakproof(get_opcode(expr->opno)))
 			return false;
 
-		/* Check (Var op Const) or (Const op Var) clauses by recursing. */
-		if (IsA(clause_expr, Var))
-			return statext_is_compatible_clause_internal(root, clause_expr,
-														 relid, attnums, exprs);
+		/*
+		 * There's always at least one expression, otherwise the clause would
+		 * not be considered compatible.
+		 */
+		Assert(list_length(clause_exprs) >= 1);
+
+		/*
+		 * Check all expressions by recursing. Var expressions are handled as
+		 * a special case (to match it to attnums etc.)
+		 */
+		foreach (lc, clause_exprs)
+		{
+			Node *clause_expr = (Node *) lfirst(lc);
+
+			if (IsA(clause_expr, Var))
+			{
+				/* if the Var is incompatible, the whole clause is incompatible */
+				if (!statext_is_compatible_clause_internal(root, clause_expr,
+														   relid, attnums, exprs))
+					return false;
+			}
+			else	/* generic expression */
+				*exprs = lappend(*exprs, clause_expr);
+		}
 
-		/* Otherwise we have (Expr op Const) or (Const op Expr). */
-		*exprs = lappend(*exprs, clause_expr);
 		return true;
 	}
 
@@ -1420,15 +1446,21 @@ statext_is_compatible_clause_internal(PlannerInfo *root, Node *clause,
 		RangeTblEntry *rte = root->simple_rte_array[relid];
 		ScalarArrayOpExpr *expr = (ScalarArrayOpExpr *) clause;
 		Node	   *clause_expr;
+		List	   *clause_exprs;
 
 		/* Only expressions with two arguments are considered compatible. */
 		if (list_length(expr->args) != 2)
 			return false;
 
 		/* Check if the expression has the right shape (one Var, one Const) */
-		if (!examine_opclause_args(expr->args, &clause_expr, NULL, NULL))
+		if (!examine_opclause_args(expr->args, &clause_exprs, NULL, NULL))
 			return false;
 
+		/* There has to be one expression exactly. */
+		Assert(list_length(clause_exprs) == 1);
+
+		clause_expr = (Node *) linitial(clause_exprs);
+
 		/*
 		 * If it's not one of the supported operators ("=", "<", ">", etc.),
 		 * just ignore the clause, as it's not compatible with MCV lists.
@@ -2015,20 +2047,19 @@ statext_clauselist_selectivity(PlannerInfo *root, List *clauses, int varRelid,
  * examine_opclause_args
  *		Split an operator expression's arguments into Expr and Const parts.
  *
- * Attempts to match the arguments to either (Expr op Const) or (Const op
- * Expr), possibly with a RelabelType on top. When the expression matches this
- * form, returns true, otherwise returns false.
+ * Attempts to match the arguments to either (Expr op Const) or (Const op Expr)
+ * or (Expr op Expr), possibly with a RelabelType on top. When the expression
+ * matches this form, returns true, otherwise returns false.
  *
  * Optionally returns pointers to the extracted Expr/Const nodes, when passed
- * non-null pointers (exprp, cstp and expronleftp). The expronleftp flag
+ * non-null pointers (exprsp, cstp and expronleftp). The expronleftp flag
  * specifies on which side of the operator we found the expression node.
  */
 bool
-examine_opclause_args(List *args, Node **exprp, Const **cstp,
-					  bool *expronleftp)
+examine_opclause_args(List *args, List **exprsp, Const **cstp, bool *expronleftp)
 {
-	Node	   *expr;
-	Const	   *cst;
+	List	   *exprs = NIL;
+	Const	   *cst = NULL;
 	bool		expronleft;
 	Node	   *leftop,
 			   *rightop;
@@ -2048,22 +2079,26 @@ examine_opclause_args(List *args, Node **exprp, Const **cstp,
 
 	if (IsA(rightop, Const))
 	{
-		expr = (Node *) leftop;
+		exprs = lappend(exprs, leftop);
 		cst = (Const *) rightop;
 		expronleft = true;
 	}
 	else if (IsA(leftop, Const))
 	{
-		expr = (Node *) rightop;
+		exprs = lappend(exprs, rightop);
 		cst = (Const *) leftop;
 		expronleft = false;
 	}
 	else
-		return false;
+	{
+		exprs = lappend(exprs, leftop);
+		exprs = lappend(exprs, rightop);
+		expronleft = false;
+	}
 
 	/* return pointers to the extracted parts if requested */
-	if (exprp)
-		*exprp = expr;
+	if (exprsp)
+		*exprsp = exprs;
 
 	if (cstp)
 		*cstp = cst;
diff --git a/src/backend/statistics/mcv.c b/src/backend/statistics/mcv.c
index f10642df4f7..5806cfbb564 100644
--- a/src/backend/statistics/mcv.c
+++ b/src/backend/statistics/mcv.c
@@ -1645,78 +1645,154 @@ mcv_get_match_bitmap(PlannerInfo *root, List *clauses,
 
 			/* valid only after examine_opclause_args returns true */
 			Node	   *clause_expr;
+			Node	   *clause_expr2;
+			List	   *clause_exprs;
 			Const	   *cst;
 			bool		expronleft;
-			int			idx;
 			Oid			collid;
 
 			fmgr_info(get_opcode(expr->opno), &opproc);
 
 			/* extract the var/expr and const from the expression */
-			if (!examine_opclause_args(expr->args, &clause_expr, &cst, &expronleft))
+			if (!examine_opclause_args(expr->args, &clause_exprs, &cst, &expronleft))
 				elog(ERROR, "incompatible clause");
 
-			/* match the attribute/expression to a dimension of the statistic */
-			idx = mcv_match_expression(clause_expr, keys, exprs, &collid);
+			if (cst)	/* Expr op Const */
+			{
+				int idx;
 
-			Assert((idx >= 0) && (idx < bms_num_members(keys) + list_length(exprs)));
+				Assert(list_length(clause_exprs) == 1);
+				clause_expr = (Node *) linitial(clause_exprs);
 
-			/*
-			 * Walk through the MCV items and evaluate the current clause. We
-			 * can skip items that were already ruled out, and terminate if
-			 * there are no remaining MCV items that might possibly match.
-			 */
-			for (i = 0; i < mcvlist->nitems; i++)
-			{
-				bool		match = true;
-				MCVItem    *item = &mcvlist->items[i];
+				/* match the attribute/expression to a dimension of the statistic */
+				idx = mcv_match_expression(clause_expr, keys, exprs, &collid);
 
-				Assert(idx >= 0);
+				Assert((idx >= 0) && (idx < bms_num_members(keys) + list_length(exprs)));
 
 				/*
-				 * When the MCV item or the Const value is NULL we can treat
-				 * this as a mismatch. We must not call the operator because
-				 * of strictness.
+				 * Walk through the MCV items and evaluate the current clause. We
+				 * can skip items that were already ruled out, and terminate if
+				 * there are no remaining MCV items that might possibly match.
 				 */
-				if (item->isnull[idx] || cst->constisnull)
+				for (i = 0; i < mcvlist->nitems; i++)
 				{
-					matches[i] = RESULT_MERGE(matches[i], is_or, false);
-					continue;
+					bool		match = true;
+					MCVItem    *item = &mcvlist->items[i];
+
+					Assert(idx >= 0);
+
+					/*
+					 * When the MCV item or the Const value is NULL we can treat
+					 * this as a mismatch. We must not call the operator because
+					 * of strictness.
+					 */
+					if (item->isnull[idx] || cst->constisnull)
+					{
+						matches[i] = RESULT_MERGE(matches[i], is_or, false);
+						continue;
+					}
+
+					/*
+					 * Skip MCV items that can't change result in the bitmap. Once
+					 * the value gets false for AND-lists, or true for OR-lists,
+					 * we don't need to look at more clauses.
+					 */
+					if (RESULT_IS_FINAL(matches[i], is_or))
+						continue;
+
+					/*
+					 * We don't store collations used to build the statistics, but
+					 * we can use the collation for the attribute itself, as
+					 * stored in varcollid. We do reset the statistics after a
+					 * type change (including collation change), so this is OK.
+					 * For expressions, we use the collation extracted from the
+					 * expression itself.
+					 */
+					if (expronleft)
+						match = DatumGetBool(FunctionCall2Coll(&opproc,
+															   collid,
+															   item->values[idx],
+															   cst->constvalue));
+					else
+						match = DatumGetBool(FunctionCall2Coll(&opproc,
+															   collid,
+															   cst->constvalue,
+															   item->values[idx]));
+
+					/* update the match bitmap with the result */
+					matches[i] = RESULT_MERGE(matches[i], is_or, match);
 				}
+			}
+			else	/* Expr op Expr */
+			{
+				int			idx;
+				int			idx2;
+
+				Assert(list_length(clause_exprs) == 2);
+
+				clause_expr = (Node *) linitial(clause_exprs);
+				clause_expr2 = (Node *) lsecond(clause_exprs);
+
+				Assert(clause_expr2);
+				Assert(!expronleft);
 
 				/*
-				 * Skip MCV items that can't change result in the bitmap. Once
-				 * the value gets false for AND-lists, or true for OR-lists,
-				 * we don't need to look at more clauses.
+				 * Match the expressions to a dimension of the statistic.
+				 *
+				 * XXX Can the collations differ?
 				 */
-				if (RESULT_IS_FINAL(matches[i], is_or))
-					continue;
+				idx = mcv_match_expression(clause_expr, keys, exprs, &collid);
+				idx2 = mcv_match_expression(clause_expr2, keys, exprs, &collid);
+
+				Assert((idx >= 0) && (idx < bms_num_members(keys) + list_length(exprs)));
+				Assert((idx2 >= 0) && (idx2 < bms_num_members(keys) + list_length(exprs)));
 
 				/*
-				 * First check whether the constant is below the lower
-				 * boundary (in that case we can skip the bucket, because
-				 * there's no overlap).
-				 *
-				 * We don't store collations used to build the statistics, but
-				 * we can use the collation for the attribute itself, as
-				 * stored in varcollid. We do reset the statistics after a
-				 * type change (including collation change), so this is OK.
-				 * For expressions, we use the collation extracted from the
-				 * expression itself.
+				 * Walk through the MCV items and evaluate the current clause.
+				 * We can skip items that were already ruled out, and
+				 * terminate if there are no remaining MCV items that might
+				 * possibly match.
 				 */
-				if (expronleft)
+				for (i = 0; i < mcvlist->nitems; i++)
+				{
+					bool		match = true;
+					MCVItem    *item = &mcvlist->items[i];
+
+					/*
+					 * When either of the MCV items is NULL we can treat this
+					 * as a mismatch. We must not call the operator because
+					 * of strictness.
+					 */
+					if (item->isnull[idx] || item->isnull[idx2])
+					{
+						matches[i] = RESULT_MERGE(matches[i], is_or, false);
+						continue;
+					}
+
+					/*
+					 * Skip MCV items that can't change result in the bitmap.
+					 * Once the value gets false for AND-lists, or true for
+					 * OR-lists, we don't need to look at more clauses.
+					 */
+					if (RESULT_IS_FINAL(matches[i], is_or))
+						continue;
+
+					/*
+					 * We don't store collations used to build the statistics,
+					 * but we can use the collation for the attribute itself,
+					 * as stored in varcollid. We do reset the statistics after
+					 * a type change (including collation change), so this is
+					 * OK. We may need to relax this after allowing extended
+					 * statistics on expressions.
+					 */
 					match = DatumGetBool(FunctionCall2Coll(&opproc,
 														   collid,
 														   item->values[idx],
-														   cst->constvalue));
-				else
-					match = DatumGetBool(FunctionCall2Coll(&opproc,
-														   collid,
-														   cst->constvalue,
-														   item->values[idx]));
+														   item->values[idx2]));
 
-				/* update the match bitmap with the result */
-				matches[i] = RESULT_MERGE(matches[i], is_or, match);
+					/* update the match bitmap with the result */
+					matches[i] = RESULT_MERGE(matches[i], is_or, match);
+				}
 			}
 		}
 		else if (IsA(clause, ScalarArrayOpExpr))
@@ -1726,6 +1802,7 @@ mcv_get_match_bitmap(PlannerInfo *root, List *clauses,
 
 			/* valid only after examine_opclause_args returns true */
 			Node	   *clause_expr;
+			List	   *clause_exprs;
 			Const	   *cst;
 			bool		expronleft;
 			Oid			collid;
@@ -1743,11 +1820,14 @@ mcv_get_match_bitmap(PlannerInfo *root, List *clauses,
 			fmgr_info(get_opcode(expr->opno), &opproc);
 
 			/* extract the var/expr and const from the expression */
-			if (!examine_opclause_args(expr->args, &clause_expr, &cst, &expronleft))
+			if (!examine_opclause_args(expr->args, &clause_exprs, &cst, &expronleft))
 				elog(ERROR, "incompatible clause");
 
 			/* ScalarArrayOpExpr has the Var always on the left */
 			Assert(expronleft);
+			Assert(list_length(clause_exprs) == 1);
+
+			clause_expr = (Node *) linitial(clause_exprs);
 
 			/* XXX what if (cst->constisnull == NULL)? */
 			if (!cst->constisnull)
diff --git a/src/include/statistics/extended_stats_internal.h b/src/include/statistics/extended_stats_internal.h
index 71f852c157b..bd4bf7fdf14 100644
--- a/src/include/statistics/extended_stats_internal.h
+++ b/src/include/statistics/extended_stats_internal.h
@@ -99,8 +99,8 @@ extern SortItem *build_sorted_items(StatsBuildData *data, int *nitems,
 									MultiSortSupport mss,
 									int numattrs, AttrNumber *attnums);
 
-extern bool examine_opclause_args(List *args, Node **exprp,
-								  Const **cstp, bool *expronleftp);
+extern bool examine_opclause_args(List *args, List **exprs, Const **cstp,
+								  bool *expronleftp);
 
 extern Selectivity mcv_combine_selectivities(Selectivity simple_sel,
 											 Selectivity mcv_sel,
diff --git a/src/test/regress/expected/stats_ext.out b/src/test/regress/expected/stats_ext.out
index fea3d18442c..73ba59d75f1 100644
--- a/src/test/regress/expected/stats_ext.out
+++ b/src/test/regress/expected/stats_ext.out
@@ -2246,37 +2246,37 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE a >= a');
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE a = c');
  estimated | actual 
 -----------+--------
-        25 |   1250
+      1250 |   1250
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE a != c');
  estimated | actual 
 -----------+--------
-      4975 |   3750
+      3750 |   3750
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE a > c');
  estimated | actual 
 -----------+--------
-      1667 |   3750
+      3750 |   3750
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE a >= c');
  estimated | actual 
 -----------+--------
-      1667 |   5000
+      5000 |   5000
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE a < c');
  estimated | actual 
 -----------+--------
-      1667 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE a <= c');
  estimated | actual 
 -----------+--------
-      1667 |   1250
+      1250 |   1250
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE a IN (1, 2, 51, 52) AND b IN ( ''1'', ''2'')');
@@ -3366,7 +3366,7 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE a IS NULL AND
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b = d');
  estimated | actual 
 -----------+--------
-        25 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b = d AND a > a');
@@ -3396,31 +3396,31 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b = d AND NOT
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b = d OR a > a');
  estimated | actual 
 -----------+--------
-        25 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b = d OR a < a');
  estimated | actual 
 -----------+--------
-        25 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b = d OR a = a');
  estimated | actual 
 -----------+--------
-        25 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b = d OR NOT a = a');
  estimated | actual 
 -----------+--------
-        25 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b != d');
  estimated | actual 
 -----------+--------
-      4975 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b != d AND a > a');
@@ -3450,31 +3450,31 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b != d AND NOT
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b != d OR a > a');
  estimated | actual 
 -----------+--------
-      4975 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b != d OR a < a');
  estimated | actual 
 -----------+--------
-      4975 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b != d OR a = a');
  estimated | actual 
 -----------+--------
-      4975 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b != d OR NOT a = a');
  estimated | actual 
 -----------+--------
-      4975 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b < d');
  estimated | actual 
 -----------+--------
-      1667 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b < d AND a > a');
@@ -3504,31 +3504,31 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b < d AND NOT
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b < d OR a > a');
  estimated | actual 
 -----------+--------
-      1667 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b < d OR a < a');
  estimated | actual 
 -----------+--------
-      1667 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b < d OR a = a');
  estimated | actual 
 -----------+--------
-      1667 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b < d OR NOT a = a');
  estimated | actual 
 -----------+--------
-      1667 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b <= d');
  estimated | actual 
 -----------+--------
-      1667 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b <= d AND a > a');
@@ -3558,31 +3558,31 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b <= d AND NOT
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b <= d OR a > a');
  estimated | actual 
 -----------+--------
-      1667 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b <= d OR a < a');
  estimated | actual 
 -----------+--------
-      1667 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b <= d OR a = a');
  estimated | actual 
 -----------+--------
-      1667 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b <= d OR NOT a = a');
  estimated | actual 
 -----------+--------
-      1667 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b > d');
  estimated | actual 
 -----------+--------
-      1667 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b > d AND a > a');
@@ -3612,31 +3612,31 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b > d AND NOT
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b > d OR a > a');
  estimated | actual 
 -----------+--------
-      1667 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b > d OR a < a');
  estimated | actual 
 -----------+--------
-      1667 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b > d OR a = a');
  estimated | actual 
 -----------+--------
-      1667 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b > d OR NOT a = a');
  estimated | actual 
 -----------+--------
-      1667 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b >= d');
  estimated | actual 
 -----------+--------
-      1667 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b >= d AND a > a');
@@ -3666,25 +3666,25 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b >= d AND NOT
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b >= d OR a > a');
  estimated | actual 
 -----------+--------
-      1667 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b >= d OR a < a');
  estimated | actual 
 -----------+--------
-      1667 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b >= d OR a = a');
  estimated | actual 
 -----------+--------
-      1667 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b >= d OR NOT a = a');
  estimated | actual 
 -----------+--------
-      1667 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b = d AND c > c');
@@ -3702,7 +3702,7 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b = d AND c <
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b = d AND c = c');
  estimated | actual 
 -----------+--------
-        12 |   2500
+      1250 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b = d AND NOT c = c');
@@ -3714,31 +3714,31 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b = d AND NOT
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b = d OR c > c');
  estimated | actual 
 -----------+--------
-        25 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b = d OR c < c');
  estimated | actual 
 -----------+--------
-        25 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b = d OR c = c');
  estimated | actual 
 -----------+--------
-      2513 |   2500
+      3750 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b = d OR NOT c = c');
  estimated | actual 
 -----------+--------
-        25 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b != d');
  estimated | actual 
 -----------+--------
-      4975 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b != d AND c > c');
@@ -3756,7 +3756,7 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b != d AND c <
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b != d AND c = c');
  estimated | actual 
 -----------+--------
-      2488 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b != d AND NOT c = c');
@@ -3768,31 +3768,31 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b != d AND NOT
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b != d OR c > c');
  estimated | actual 
 -----------+--------
-      4975 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b != d OR c < c');
  estimated | actual 
 -----------+--------
-      4975 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b != d OR c = c');
  estimated | actual 
 -----------+--------
-      4988 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b != d OR NOT c = c');
  estimated | actual 
 -----------+--------
-      4975 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b < d');
  estimated | actual 
 -----------+--------
-      1667 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b < d AND c > c');
@@ -3810,7 +3810,7 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b < d AND c <
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b < d AND c = c');
  estimated | actual 
 -----------+--------
-       833 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b < d AND NOT c = c');
@@ -3822,31 +3822,31 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b < d AND NOT
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b < d OR c > c');
  estimated | actual 
 -----------+--------
-      1667 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b < d OR c < c');
  estimated | actual 
 -----------+--------
-      1667 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b < d OR c = c');
  estimated | actual 
 -----------+--------
-      3333 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b < d OR NOT c = c');
  estimated | actual 
 -----------+--------
-      1667 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b <= d');
  estimated | actual 
 -----------+--------
-      1667 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b <= d AND c > c');
@@ -3864,7 +3864,7 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b <= d AND c <
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b <= d AND c = c');
  estimated | actual 
 -----------+--------
-       833 |   2500
+      1250 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b <= d AND NOT c = c');
@@ -3876,31 +3876,31 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b <= d AND NOT
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b <= d OR c > c');
  estimated | actual 
 -----------+--------
-      1667 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b <= d OR c < c');
  estimated | actual 
 -----------+--------
-      1667 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b <= d OR c = c');
  estimated | actual 
 -----------+--------
-      3333 |   2500
+      3750 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b <= d OR NOT c = c');
  estimated | actual 
 -----------+--------
-      1667 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b > d');
  estimated | actual 
 -----------+--------
-      1667 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b > d AND c > c');
@@ -3918,7 +3918,7 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b > d AND c <
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b > d AND c = c');
  estimated | actual 
 -----------+--------
-       833 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b > d AND NOT c = c');
@@ -3930,31 +3930,31 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b > d AND NOT
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b > d OR c > c');
  estimated | actual 
 -----------+--------
-      1667 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b > d OR c < c');
  estimated | actual 
 -----------+--------
-      1667 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b > d OR c = c');
  estimated | actual 
 -----------+--------
-      3333 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b > d OR NOT c = c');
  estimated | actual 
 -----------+--------
-      1667 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b >= d');
  estimated | actual 
 -----------+--------
-      1667 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b >= d AND c > c');
@@ -3972,7 +3972,7 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b >= d AND c <
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b >= d AND c = c');
  estimated | actual 
 -----------+--------
-       833 |   2500
+      1250 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b >= d AND NOT c = c');
@@ -3984,25 +3984,25 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b >= d AND NOT
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b >= d OR c > c');
  estimated | actual 
 -----------+--------
-      1667 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b >= d OR c < c');
  estimated | actual 
 -----------+--------
-      1667 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b >= d OR c = c');
  estimated | actual 
 -----------+--------
-      3333 |   2500
+      3750 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists WHERE b >= d OR NOT c = c');
  estimated | actual 
 -----------+--------
-      1667 |   2500
+      2500 |   2500
 (1 row)
 
 -- mcv with pass-by-ref fixlen types, e.g. uuid
@@ -4345,37 +4345,37 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_bool WHERE NOT a >=
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_bool WHERE NOT a = b');
  estimated | actual 
 -----------+--------
-      9950 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_bool WHERE NOT a != b');
  estimated | actual 
 -----------+--------
-        50 |   7500
+      7500 |   7500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_bool WHERE NOT a < b');
  estimated | actual 
 -----------+--------
-      3333 |  10000
+     10000 |  10000
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_bool WHERE NOT a <= b');
  estimated | actual 
 -----------+--------
-      3333 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_bool WHERE NOT a > b');
  estimated | actual 
 -----------+--------
-      3333 |   7500
+      7500 |   7500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_bool WHERE NOT a >= b');
  estimated | actual 
 -----------+--------
-      3333 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_bool WHERE NOT a = a AND b = b');
@@ -4417,43 +4417,43 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_bool WHERE NOT a >=
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_bool WHERE NOT a = b AND b = c');
  estimated | actual 
 -----------+--------
-        50 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_bool WHERE NOT a != b AND b != c');
  estimated | actual 
 -----------+--------
-        50 |   1250
+      1250 |   1250
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_bool WHERE NOT a < b AND b < c');
  estimated | actual 
 -----------+--------
-      1111 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_bool WHERE NOT a <= b AND b <= c');
  estimated | actual 
 -----------+--------
-      1111 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_bool WHERE NOT a > b AND b > c');
  estimated | actual 
 -----------+--------
-      1111 |   1250
+      1250 |   1250
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_bool WHERE NOT a >= b AND b >= c');
  estimated | actual 
 -----------+--------
-      1111 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_bool WHERE NOT a = c AND b = b');
  estimated | actual 
 -----------+--------
-      9950 |   3750
+      3750 |   3750
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_bool WHERE NOT a != c AND b != b');
@@ -4477,7 +4477,7 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_bool WHERE NOT a <=
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_bool WHERE NOT a > a AND b > c');
  estimated | actual 
 -----------+--------
-      3333 |   1250
+      1250 |   1250
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_bool WHERE NOT a >= a AND b >= c');
@@ -4617,13 +4617,13 @@ DROP TABLE mcv_lists_partial;
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_bool WHERE NOT a = b');
  estimated | actual 
 -----------+--------
-      9950 |   2500
+      2500 |   2500
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_bool WHERE NOT a = b AND b = c');
  estimated | actual 
 -----------+--------
-        50 |   2500
+      2500 |   2500
 (1 row)
 
 -- check the ability to use multiple MCV lists
@@ -4854,55 +4854,55 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_multi WHERE a = 0 OR
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_multi WHERE a = b AND c = d');
  estimated | actual 
 -----------+--------
-         1 |   5000
+      5000 |   5000
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_multi WHERE a != b AND c != d');
  estimated | actual 
 -----------+--------
-      4950 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_multi WHERE a < b AND c > d');
  estimated | actual 
 -----------+--------
-       556 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_multi WHERE a <= b AND c >= d');
  estimated | actual 
 -----------+--------
-       556 |   5000
+      5000 |   5000
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_multi WHERE a = b OR c = d');
  estimated | actual 
 -----------+--------
-        50 |   5000
+      5000 |   5000
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_multi WHERE a != b OR c != d');
  estimated | actual 
 -----------+--------
-      5000 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_multi WHERE a < b OR c > d');
  estimated | actual 
 -----------+--------
-      2778 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_multi WHERE a <= b OR c >= d');
  estimated | actual 
 -----------+--------
-      2778 |   5000
+      5000 |   5000
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_multi WHERE a = a AND c = d');
  estimated | actual 
 -----------+--------
-        25 |   5000
+      5000 |   5000
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_multi WHERE a != b AND c != c');
@@ -4920,7 +4920,7 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_multi WHERE a < a AN
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_multi WHERE a <= b AND d >= c');
  estimated | actual 
 -----------+--------
-       556 |   5000
+      5000 |   5000
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_multi WHERE a = a OR c = d');
@@ -4932,43 +4932,43 @@ SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_multi WHERE a = a OR
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_multi WHERE a != b OR c != c');
  estimated | actual 
 -----------+--------
-      4975 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_multi WHERE a < a OR c > d');
  estimated | actual 
 -----------+--------
-      1667 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_multi WHERE a <= b OR d >= c');
  estimated | actual 
 -----------+--------
-      2778 |   5000
+      5000 |   5000
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_multi WHERE a < b');
  estimated | actual 
 -----------+--------
-      1667 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_multi WHERE a > b');
  estimated | actual 
 -----------+--------
-      1667 |      0
+         1 |      0
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_multi WHERE a <= b');
  estimated | actual 
 -----------+--------
-      1667 |   5000
+      5000 |   5000
 (1 row)
 
 SELECT * FROM check_estimated_rows('SELECT * FROM mcv_lists_multi WHERE a >= b');
  estimated | actual 
 -----------+--------
-      1667 |   5000
+      5000 |   5000
 (1 row)
 
 DROP TABLE mcv_lists_multi;
-- 
2.34.3

